from coalib.bears.requirements.PackageRequirement import PackageRequirement
from coalib.misc.Shell import call_without_output

from coala_utils.string_processing import escape


class JuliaRequirement(PackageRequirement):
    """
    This class is a subclass of ``PackageRequirement``. It specifies the proper
    type for ``julia`` packages automatically and provide a function to check
    for the requirement.
    """

    def __init__(self, package, version=''):
        """
        Constructs a new ``JuliaRequirement``, using the ``PackageRequirement``
        constructor.

        >>> pr = JuliaRequirement('Lint', '19.2')
        >>> pr.type
        'julia'
        >>> pr.package
        'Lint'
        >>> pr.version
        '19.2'

        :param package: A string with the name of the package to be installed.
        :param version: A version string. Leave empty to specify latest version.
        """
        PackageRequirement.__init__(self, 'julia', package, version)

    def is_installed(self):
        """
        Checks if the dependency is installed.

        :return: ``True`` if dependency is installed, ``False`` otherwise.
        """
        # We need to check explicitly if `nothing` is returned, as this happens
        # when the package is *registered*, but *not installed*. If it's not
        # even registered, julia will throw an exception which lets julia exit
        # with an error code different from 0.
        code = 'Pkg.installed("{}")==nothing?exit(1):exit(0)'.format(
            escape(self.package, '\\"'))
        args = ('julia', '-e', code)
        return not call_without_output(args)



from coalib.bears.requirements.PackageRequirement import PackageRequirement
from coalib.misc.Shell import call_without_output
import platform


class GemRequirement(PackageRequirement):
    """
    This class is a subclass of ``PackageRequirement``. It specifies the proper
    type for ``ruby`` packages automatically and provide a function to check
    for the requirement.
    """

    def __init__(self, package, version='', require=''):
        """
        Constructs a new ``GemRequirement``, using the ``PackageRequirement``
        constructor.

        >>> pr = GemRequirement('setuptools', '19.2', 'flag')
        >>> pr.type
        'gem'
        >>> pr.package
        'setuptools'
        >>> pr.version
        '19.2'
        >>> pr.require
        'flag'

        :param package: A string with the name of the package to be installed.
        :param version: A version string. Leave empty to specify latest version.
        :param require: A string that specifies any additional flags, that
                        would be used with ``require``.
        """
        PackageRequirement.__init__(self, 'gem', package, version)
        self.require = require

    def is_installed(self):
        """
        Checks if the dependency is installed.

        :param return: True if dependency is installed, false otherwise.
        """
        cmd = ['gem', 'list', '-i', self.package]
        if platform.system() == 'Windows':  # pragma: no cover
            cmd = ['cmd', '/c'] + cmd
        return not call_without_output(cmd)



from coalib.results.Result import Result
from coalib.results.result_actions.ResultAction import ResultAction


class PrintMoreInfoAction(ResultAction):

    @staticmethod
    def is_applicable(result, original_file_dict, file_diff_dict):
        return isinstance(result, Result) and result.additional_info != ''

    def apply(self, result, original_file_dict, file_diff_dict):
        """
        Print more info
        """
        print(result.additional_info)

        return file_diff_dict



from coalib.bears.requirements.PackageRequirement import PackageRequirement
from coalib.misc.Shell import run_shell_command


class RscriptRequirement(PackageRequirement):
    """
    This class is a subclass of ``PackageRequirement``. It specifies the proper
    type for ``R`` packages automatically and provide a function to check
    for the requirement.
    """

    def __init__(self, package, version='', flag='', repo=''):
        """
        Constructs a new ``RscriptRequirement``, using the
        ``PackageRequirement`` constructor.

        >>> pr = RscriptRequirement(
        ...         'formatR', version='1.4', flag='-e',
        ...         repo="http://cran.rstudio.com")
        >>> pr.type
        'R'
        >>> pr.package
        'formatR'
        >>> pr.version
        '1.4'
        >>> pr.flag
        '-e'
        >>> pr.repo
        'http://cran.rstudio.com'

        :param package: A string with the name of the package to be installed.
        :param version: A version string. Leave empty to specify latest version.
        :param flag:    A string that specifies any additional flags, that
                        are passed to the type.
        :param repo:    The repository from which the package to be installed is
                        from.
        """
        PackageRequirement.__init__(self, 'R', package, version)
        self.flag = flag
        self.repo = repo

    def is_installed(self):
        """
        Checks if the dependency is installed.

        :param return: True if dependency is installed, false otherwise.
        """
        return True if run_shell_command(
                ('R -e \'library(\"{}\", quietly=TRUE)\''
                 .format(self.package)))[1] is '' else False



from coala_utils.decorators import generate_eq, generate_repr


@generate_eq('type', 'package', 'version')
@generate_repr()
class PackageRequirement:
    """
    This class helps keeping track of bear requirements. It should simply
    be appended to the REQUIREMENTS tuple inside the Bear class.

    Two ``PackageRequirements`` should always be equal if they have the same
    type, package and version:

    >>> pr1 = PackageRequirement('pip', 'coala_decorators', '0.1.0')
    >>> pr2 = PackageRequirement('pip', 'coala_decorators', '0.1.0')
    >>> pr1 == pr2
    True
    """

    def __init__(self, type: str, package: str, version=''):
        """
        Constructs a new ``PackageRequirement``.

        >>> pr = PackageRequirement('pip', 'colorama', '0.1.0')
        >>> pr.type
        'pip'
        >>> pr.package
        'colorama'
        >>> pr.version
        '0.1.0'

        :param type:    A string with the name of the manager (pip, npm, etc).
        :param package: A string with the name of the package to be installed.
        :param version: A version string. Leave empty to specify latest version.
        """
        self.type = type
        self.package = package
        self.version = version

    def is_installed(self):
        """
        Check if the requirement is satisfied.

        >>> PackageRequirement('pip', \
                               'coala_decorators', \
                               '0.2.1').is_installed()
        Traceback (most recent call last):
        ...
        NotImplementedError

        :return: Returns True if satisfied, False if not.
        """
        raise NotImplementedError



import re


def to_camelcase(string):
    """
    Converts the given string to camel-case.

    >>> to_camelcase('Hello_world')
    'helloWorld'
    >>> to_camelcase('__Init__file__')
    '__initFile__'
    >>> to_camelcase('')
    ''
    >>> to_camelcase('alreadyCamelCase')
    'alreadyCamelCase'
    >>> to_camelcase('   string')
    '___string'

    :param string: The string to convert.
    :return:       The camel-cased string.
    """
    string = re.sub('(\s)',
                    lambda match: '_',
                    string)
    string = re.sub('^(_*)(.)',
                    lambda match: match.group(1) + match.group(2).lower(),
                    string)
    return re.sub('(?<=[^_])_+([^_])',
                  lambda match: match.group(1).upper(),
                  string)


def to_pascalcase(string):
    """
    Converts the given to string pascal-case.

    >>> to_pascalcase('hello_world')
    'HelloWorld'
    >>> to_pascalcase('__init__file__')
    '__InitFile__'
    >>> to_pascalcase('')
    ''
    >>> to_pascalcase('AlreadyPascalCase')
    'AlreadyPascalCase'
    >>> to_pascalcase('   string')
    '___String'

    :param string: The string to convert.
    :return:       The pascal-cased string.
    """
    string = re.sub('(\s)',
                    lambda match: '_',
                    string)
    string = re.sub('^(_*)(.)',
                    lambda match: match.group(1) + match.group(2).upper(),
                    string)
    return re.sub('(?<=[^_])_+([^_])',
                  lambda match: match.group(1).upper(),
                  string)


def to_snakecase(string):
    """
    Converts the given string to snake-case.

    >>> to_snakecase('HelloWorld')
    'hello_world'
    >>> to_snakecase('__Init__File__')
    '__init_file__'
    >>> to_snakecase('')
    ''
    >>> to_snakecase('already_snake_case')
    'already_snake_case'
    >>> to_snakecase('   string  ')
    '___string__'
    >>> to_snakecase('ABCde.F.G..H..IH')
    'a_b_cde.f.g..h..i_h'

    :param string: The string to convert.
    :return:       The snake-cased string.
    """
    string = re.sub('(\s)',
                    lambda match: '_',
                    string)
    string = re.sub('^(_*)([^_])',
                    lambda match: match.group(1) + match.group(2).lower(),
                    string)
    string = re.sub('(\w*)([.]+)([A-Z])',
                    lambda match: (match.group(1) + match.group(2) +
                                   match.group(3).lower()),
                    string)
    string = re.sub('(?<=[^_])_+([^_])',
                    lambda match: '_' + match.group(1).lower(),
                    string)
    return re.sub('[A-Z]',
                  lambda match: '_' + match.group(0).lower(),
                  string)


def to_spacecase(string):
    """
    Converts the given string to space-case.

    >>> to_spacecase('helloWorld')
    'Hello World'
    >>> to_spacecase('__Init__File__')
    'Init File'
    >>> to_spacecase('')
    ''
    >>> to_spacecase('Already Space Case')
    'Already Space Case'
    >>> to_spacecase('  string  ')
    'String'

    :param string: The string to convert.
    :return:       The space-cased string.
    """
    string = re.sub('(_)',
                    ' ',
                    string)
    string = re.sub('^(\s*)(.)',
                    lambda match: match.group(2).upper(),
                    string)
    string = re.sub('(\s*)$',
                    '',
                    string)
    string = re.sub('(?<=[^\s])\s+([^\s])',
                    lambda match: ' ' + match.group(1).upper(),
                    string)
    return re.sub('(?<=[^\s])([A-Z])',
                  lambda match: ' ' + match.group(1),
                  string)



import shutil
from os.path import isfile
from os import remove

from coalib.results.result_actions.ShowPatchAction import ShowPatchAction
from coalib.results.result_actions.ResultAction import ResultAction


class ApplyPatchAction(ResultAction):

    SUCCESS_MESSAGE = 'Patch applied successfully.'

    is_applicable = staticmethod(ShowPatchAction.is_applicable)

    def apply(self,
              result,
              original_file_dict,
              file_diff_dict,
              no_orig: bool=False):
        """
        Apply patch

        :param no_orig: Whether or not to create .orig backup files
        """
        for filename in result.diffs:
            pre_patch_filename = filename
            if filename in file_diff_dict:
                diff = file_diff_dict[filename]
                pre_patch_filename = (diff.rename
                                      if diff.rename is not False
                                      else filename)
                file_diff_dict[filename] += result.diffs[filename]
            else:
                file_diff_dict[filename] = result.diffs[filename]

                # Backup original file, only if there was no previous patch
                # from this run though!
                if not no_orig and isfile(pre_patch_filename):
                    shutil.copy2(pre_patch_filename,
                                 pre_patch_filename + '.orig')

            diff = file_diff_dict[filename]
            if diff.delete or diff.rename:
                if isfile(pre_patch_filename):
                    remove(pre_patch_filename)
            if not diff.delete:
                new_filename = (diff.rename
                                if diff.rename is not False
                                else filename)
                with open(new_filename, mode='w', encoding='utf-8') as file:
                    file.writelines(diff.modified)

        return file_diff_dict
